"""
Template Helpers for AlgoBOWL
-----------------------------

This module is exposed in all templates under the name ``h``.
"""

import datetime

import tg


def ftime(dt, duration=None):
    """Format a ``datetime`` object as HTML for view on the page.

    Localtime is normally handled by the browser.  We create <time> elements
    here, and allow client-side javascript to re-format the displayed time.

    Args:
        dt: A datetime object to format.
        duration: An optional datetime or timedelta for the ending time.

    Returns:
        A HTML string.
    """
    fallback_datetime_format = tg.config.get(
        "locale.fallback_datetime_format", "%B %-d, %Y at %-I:%M %p"
    )

    fallback = dt.strftime(fallback_datetime_format)
    timestamp = dt.astimezone().isoformat()

    result = f'<time datetime="{timestamp}">{fallback}</time>'
    if duration:
        if isinstance(duration, datetime.timedelta):
            # convert a duration to an end date
            duration = dt + duration
        if duration < dt:
            raise ValueError("Duration cannot be before the specified time")
        result += f" to {ftime(duration)}"
    return result


def url(path: str) -> str:
    try:
        from algobowl.lib import algocdn  # noqa: PLC0415
    except ImportError:
        pass
    else:
        try:
            return algocdn.url_map[path]
        except KeyError:
            pass
    return tg.url(path)



import json
import sys
from collections.abc import Iterable

import click
import requests

import algobowl.cli.formatter as fmt
from algobowl.lib import constants


def create_group(
    cli,
    competition_id: int,
    name: str = "",
    usernames: Iterable[str] = (),
    incognito: bool = False,
    benchmark: bool = False,
):
    data = {
        "competition_id": str(competition_id),
        "users": ",".join(usernames),
    }
    if name:
        data["name"] = name
    if incognito:
        data["incognito"] = "true"
    if benchmark:
        data["benchmark"] = "true"
    r = cli.session.post(cli.config.get_url("/setup/create_group.json"), data=data)
    r.raise_for_status()
    return r.json()


@click.group(
    help="Administration",
    epilog="Note: --sudo is implicit for this command group.",
)
@click.pass_obj
def admin(cli):
    # Give implicit --sudo for admin commands.
    cli.session.headers[constants.SUDO_HEADER] = "True"


@admin.command(help="Setup a new competition")
@click.option(
    "--name",
    type=str,
    help="Name of the competition",
    required=True,
)
@click.option(
    "--problem",
    type=str,
    help="Name of the problem",
    required=True,
)
@click.option(
    "--input-upload-due",
    type=str,
    help="Due date of input upload in ISO 8601 format (please include timezone!)",
    required=True,
)
@click.option(
    "--output-upload-hours",
    type=int,
    default=48,
    help="Number of hours for output upload.",
)
@click.option(
    "--verification-hours",
    type=int,
    default=24,
    help="Number of hours for verification.",
)
@click.option(
    "--resolution-hours",
    type=int,
    default=24,
    help="Number of hours for resolution.",
)
@click.option(
    "--ov-hours",
    type=int,
    default=72,
    help="Number of hours for open verification and evaluation.",
)
@click.pass_obj
def setup_competition(
    cli,
    name,
    problem,
    input_upload_due,
    output_upload_hours,
    verification_hours,
    resolution_hours,
    ov_hours,
):
    data = {
        "name": name,
        "problem": problem,
        "input_upload_ends": input_upload_due,
        "output_upload_hours": output_upload_hours,
        "verification_hours": verification_hours,
        "resolution_hours": resolution_hours,
        "ov_hours": ov_hours,
    }
    r = cli.session.post(cli.config.get_url("/setup/setup_competition.json"), data=data)
    r.raise_for_status()
    cli.formatter.dump_table([r.json()])


@admin.command(name="create-group", help="Create a new group")
@click.option(
    "--incognito/--no-incognito", "-i", help="Enable if the group should be incognito"
)
@click.option(
    "--benchmark/--no-benchmark",
    "-b",
    help="Enable if the group should be a benchmark group",
)
@click.option("--name", "-n", type=str, help="Group Name")
@click.option(
    "--competition", "-c", type=int, help="Competition ID for this group", required=True
)
@click.option("--user", "-u", type=str, help="Add user", multiple=True)
@click.pass_obj
def create_group_(cli, name, competition, user, incognito, benchmark):
    cli.formatter.dump_table(
        [
            create_group(
                cli,
                competition,
                name=name,
                usernames=user,
                incognito=incognito,
                benchmark=benchmark,
            )
        ]
    )


@admin.command(name="create-groups", help="Create groups from JSON file")
@click.option(
    "--competition", "-c", type=int, help="Competition ID for this group", required=True
)
@click.argument("json_file", type=click.Path(exists=True))
@click.pass_obj
def create_groups(cli, competition, json_file):
    with open(json_file, encoding="utf-8") as f:
        groups = json.load(f)
    failed_groups = []
    with click.progressbar(groups, label="Create groups") as bar:
        for group in bar:
            try:
                create_group(cli, competition, usernames=group)
            except (
                requests.exceptions.HTTPError,
                requests.exceptions.JSONDecodeError,
            ) as e:
                failed_groups.append((group, e))
    click.echo(
        f"Successfully created {len(groups) - len(failed_groups)} groups.", err=True
    )
    if failed_groups:
        fmt.err(f"Failed to create {len(failed_groups)} groups:")
        for group, exn in failed_groups:
            fmt.err(f"{json.dumps(group)}: {exn}")
        fmt.err("See the server logs for more information.")
        sys.exit(1)



"""The base Controller API."""

import tg

__all__ = ["BaseController"]


class BaseController(tg.TGController):
    """
    Base class for the controllers in the application.

    Your web application should have one of these. The root of
    your application is used to compute URLs used by your app.

    """

    def __call__(self, environ, context):
        """Invoke the Controller"""
        # TGController.__call__ dispatches to the Controller method
        # the request is routed to.

        tg.tmpl_context.identity = tg.request.identity
        environ["is_admin"] = tg.predicates.has_permission("admin").is_met(environ)
        return super().__call__(environ, context)



import datetime
import io
import random

import tg
from depot.io.utils import FileIntent

import algobowl.lib.problem as problemlib
from algobowl import model
from algobowl.lib import base


def get_user(username):
    user = (
        model.DBSession.query(model.User)
        .filter(model.User.username == username)
        .one_or_none()
    )
    if user:
        return user

    mpapi = tg.request.environ["repoze.who.plugins"]["mpapi"]
    return mpapi.new_user_by_username(username)


def generate_default_input(
    problem: problemlib.Problem,
    group: model.Group,
) -> model.Input:
    reformatted_contents = io.StringIO()
    iput = problem.generate_input(random.SystemRandom())
    iput.write(reformatted_contents)

    input_file = FileIntent(
        io.BytesIO(reformatted_contents.getvalue().encode("utf-8")),
        f"input_group{group.id}.txt",
        "application/octet-stream",
    )
    result = model.Input(data=input_file, group=group, is_default=True)
    model.DBSession.add(result)
    return result


class SetupController(base.BaseController):
    allow_only = tg.predicates.has_permission("admin")

    @tg.expose("json")
    def create_group(
        self,
        competition_id=None,
        name=None,
        users="",
        incognito=False,
        benchmark=False,
    ):
        competition = (
            model.DBSession.query(model.Competition)
            .filter(model.Competition.id == int(competition_id))
            .one()
        )
        problem = problemlib.load_problem(competition)
        users = [get_user(username) for username in users.split(",")]
        group = model.Group(
            users=users,
            competition=competition,
            name=name or None,
            incognito=bool(incognito),
            benchmark=bool(benchmark),
        )
        model.DBSession.add(group)
        group.input = generate_default_input(problem=problem, group=group)
        model.DBSession.flush()
        return {"group_id": group.id}

    @tg.expose("json")
    def setup_competition(
        self,
        name=None,
        problem=None,
        input_upload_ends=None,
        output_upload_hours=48,
        verification_hours=24,
        resolution_hours=24,
        ov_hours=72,
    ):
        input_upload_ends_dt = datetime.datetime.fromisoformat(input_upload_ends)
        output_upload_delta = datetime.timedelta(hours=int(output_upload_hours))
        verification_delta = datetime.timedelta(hours=int(verification_hours))
        resolution_delta = datetime.timedelta(hours=int(resolution_hours))
        ov_delta = datetime.timedelta(hours=int(ov_hours))

        output_upload_ends = input_upload_ends_dt + output_upload_delta
        verification_ends = output_upload_ends + verification_delta
        resolution_ends = verification_ends + resolution_delta
        ov_ends = resolution_ends + ov_delta

        competition = model.Competition(
            name=name,
            problem=problem,
            allow_custom_team_names=True,
            input_upload_begins=input_upload_ends_dt - datetime.timedelta(weeks=2),
            input_upload_ends=input_upload_ends_dt,
            output_upload_begins=input_upload_ends_dt,
            output_upload_ends=output_upload_ends,
            verification_begins=output_upload_ends,
            verification_ends=verification_ends,
            resolution_begins=verification_ends,
            resolution_ends=resolution_ends,
            open_verification_begins=resolution_ends,
            open_verification_ends=ov_ends,
            evaluation_begins=resolution_ends,
            evaluation_ends=ov_ends,
        )
        model.DBSession.add(competition)
        model.DBSession.flush()

        return {"competition_id": competition.id}



import dataclasses
import pathlib

import click
import requests

import algobowl.cli.config as cfg
import algobowl.cli.formatter as fmt
from algobowl.cli import admin, auth, group, problem


@dataclasses.dataclass
class CommandContext:
    config: cfg.CLIConfig = None
    session: requests.Session = None
    formatter: fmt.JsonFormatter = dataclasses.field(default_factory=fmt.TableFormatter)
    selected_group: int = -1


@click.group(help="The AlgoBOWL CLI")
@click.option("--config", type=pathlib.Path, help="Config file to use")
@click.option("--server", help="Server to use")
@click.option("--json", is_flag=True, help="Use JSON output instead of tables")
@click.option(
    "--sudo/--no-sudo",
    type=bool,
    default=False,
    help=(
        "Grant admin permissions on the server.  Doesn't do anything unless "
        "your account is actually an admin."
    ),
)
@click.pass_context
def main(ctx, server, config, json, sudo):
    ctx.ensure_object(CommandContext)

    if not config:
        config = cfg.get_default_config_path()

    ctx.obj.config = cfg.CLIConfig(path=config, server=server)
    ctx.obj.session = auth.get_session(ctx.obj.config, sudo=sudo)
    if json:
        ctx.obj.formatter = fmt.JsonFormatter()
    ctx.call_on_close(ctx.obj.config.write)


main.add_command(auth.auth)
main.add_command(cfg.config)
main.add_command(group.group)
main.add_command(admin.admin)
main.add_command(problem.problem)


if __name__ == "__main__":
    main()
